package sun.nio.ch;

import jdk.net.NetworkPermission;
import java.net.StandardProtocolFamily;
import java.net.StandardSocketOptions;
import jdk.net.SocketFlow;
import java.net.SocketOption;
import java.io.FileDescriptor;
import java.security.AccessController;
import java.util.Enumeration;
import java.security.PrivilegedAction;
import java.net.NetworkInterface;
import java.net.UnknownHostException;
import java.io.IOException;
import java.nio.channels.NotYetBoundException;
import java.nio.channels.AlreadyBoundException;
import java.nio.channels.NotYetConnectedException;
import java.nio.channels.ClosedChannelException;
import java.net.SocketException;
import java.net.InetAddress;
import java.nio.channels.UnsupportedAddressTypeException;
import java.nio.channels.UnresolvedAddressException;
import java.net.Inet6Address;
import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import sun.net.ExtendedSocketOptions;
import java.net.ProtocolFamily;

public class Net
{
    static final ProtocolFamily UNSPEC;
    private static final boolean exclusiveBind;
    private static final boolean fastLoopback;
    private static volatile boolean checkedIPv6;
    private static volatile boolean isIPv6Available;
    static final ExtendedSocketOptions extendedOptions;
    public static final int SHUT_RD = 0;
    public static final int SHUT_WR = 1;
    public static final int SHUT_RDWR = 2;
    public static final short POLLIN;
    public static final short POLLOUT;
    public static final short POLLERR;
    public static final short POLLHUP;
    public static final short POLLNVAL;
    public static final short POLLCONN;
    
    private Net() {
    }
    
    static boolean isIPv6Available() {
        if (!Net.checkedIPv6) {
            Net.isIPv6Available = isIPv6Available0();
            Net.checkedIPv6 = true;
        }
        return Net.isIPv6Available;
    }
    
    static boolean useExclusiveBind() {
        return Net.exclusiveBind;
    }
    
    static boolean canIPv6SocketJoinIPv4Group() {
        return canIPv6SocketJoinIPv4Group0();
    }
    
    static boolean canJoin6WithIPv4Group() {
        return canJoin6WithIPv4Group0();
    }
    
    public static InetSocketAddress checkAddress(SocketAddress socketAddress) {
        if (socketAddress == null) {
            throw new NullPointerException();
        }
        if (!(socketAddress instanceof InetSocketAddress)) {
            throw new UnsupportedAddressTypeException();
        }
        InetSocketAddress inetSocketAddress = (InetSocketAddress)socketAddress;
        if (inetSocketAddress.isUnresolved()) {
            throw new UnresolvedAddressException();
        }
        InetAddress address = inetSocketAddress.getAddress();
        if (!(address instanceof Inet4Address) && !(address instanceof Inet6Address)) {
            throw new IllegalArgumentException("Invalid address type");
        }
        return inetSocketAddress;
    }
    
    static InetSocketAddress asInetSocketAddress(SocketAddress socketAddress) {
        if (!(socketAddress instanceof InetSocketAddress)) {
            throw new UnsupportedAddressTypeException();
        }
        return (InetSocketAddress)socketAddress;
    }
    
    static void translateToSocketException(Exception ex) throws SocketException {
        if ((ex instanceof SocketException)) {
            throw (SocketException)ex;
        }
        if ((ex instanceof ClosedChannelException)) {
            ex = new SocketException("Socket is closed");
        }
        else if ((ex instanceof NotYetConnectedException)) {
            ex = new SocketException("Socket is not connected");
        }
        else if ((ex instanceof AlreadyBoundException)) {
            ex = new SocketException("Already bound");
        }
        else if ((ex instanceof NotYetBoundException)) {
            ex = new SocketException("Socket is not bound yet");
        }
        else if ((ex instanceof UnsupportedAddressTypeException)) {
            ex = new SocketException("Unsupported address type");
        }
        else if ((ex instanceof UnresolvedAddressException)) {
            ex = new SocketException("Unresolved address");
        }
        if (ex != ex) {
            ex.initCause(ex);
        }
        if ((ex instanceof SocketException)) {
            throw (SocketException)ex;
        }
        if ((ex instanceof RuntimeException)) {
            throw (RuntimeException)ex;
        }
        throw new Error("Untranslated exception", ex);
    }
    
    static void translateException(Exception ex, boolean b) throws IOException {
        if ((ex instanceof IOException)) {
            throw (IOException)ex;
        }
        if (b && ex instanceof UnresolvedAddressException) {
            throw new UnknownHostException();
        }
        translateToSocketException(ex);
        return;
    }
    
    static void translateException(Exception ex) throws IOException {
        translateException(ex, false);
    }
    
    static InetSocketAddress getRevealedLocalAddress(InetSocketAddress loopbackAddress) {
        SecurityManager securityManager = System.getSecurityManager();
        if (loopbackAddress == null || securityManager == null) {
            return loopbackAddress;
        }
        try {
            securityManager.checkConnect(loopbackAddress.getAddress().getHostAddress(), -1);
        }
        catch (SecurityException ex) {
            loopbackAddress = getLoopbackAddress(loopbackAddress.getPort());
        }
        return loopbackAddress;
    }
    
    static String getRevealedLocalAddressAsString(InetSocketAddress inetSocketAddress) {
        return (System.getSecurityManager() == null) ? inetSocketAddress.toString() : getLoopbackAddress(inetSocketAddress.getPort()).toString();
    }
    
    private static InetSocketAddress getLoopbackAddress(int port) {
        return new InetSocketAddress(InetAddress.getLoopbackAddress(), port);
    }
    
    static Inet4Address anyInet4Address(final NetworkInterface networkInterface) {
        return (Inet4Address)AccessController.doPrivileged(new PrivilegedAction<Inet4Address>() {
            @Override
            public Inet4Address run() {
                Enumeration<InetAddress> inetAddresses = networkInterface.getInetAddresses();
                while (inetAddresses.hasMoreElements()) {
                    InetAddress inetAddress = (InetAddress)inetAddresses.nextElement();
                    if ((inetAddress instanceof Inet4Address)) {
                        return (Inet4Address)inetAddress;
                    }
                    continue;
                }
                return null;
            }
        });
    }
    
    static int inet4AsInt(InetAddress inetAddress) {
        if ((inetAddress instanceof Inet4Address)) {
            byte[] address = inetAddress.getAddress();
            return (address[3] & 0xFF) | (address[2] << 8 & 0xFF00) | (address[1] << 16 & 0xFF0000) | (address[0] << 24 & 0xFF000000);
        }
        throw new AssertionError("Should not reach here");
    }
    
    static InetAddress inet4FromInt(int n) {
        byte[] addr = { (byte)(n >>> 24 & 0xFF), (byte)(n >>> 16 & 0xFF), (byte)(n >>> 8 & 0xFF), (byte)(n & 0xFF) };
        try {
            return InetAddress.getByAddress(addr);
        }
        catch (UnknownHostException ex) {
            throw new AssertionError("Should not reach here");
        }
    }
    
    static byte[] inet6AsByteArray(InetAddress inetAddress) {
        if ((inetAddress instanceof Inet6Address)) {
            return inetAddress.getAddress();
        }
        if ((inetAddress instanceof Inet4Address)) {
            byte[] address = inetAddress.getAddress();
            return new byte[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, address[0], address[1], address[2], address[3] };
        }
        throw new AssertionError("Should not reach here");
    }
    
    static void setSocketOption(FileDescriptor fileDescriptor, ProtocolFamily protocolFamily, SocketOption<?> socketOption, Object o) throws IOException {
        if (o == null) {
            throw new IllegalArgumentException("Invalid option value");
        }
        Class<Object> type = socketOption.type();
        if (type == SocketFlow.class) {
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                securityManager.checkPermission(new NetworkPermission("setOption.SO_FLOW_SLA"));
            }
            jdk.net.ExtendedSocketOptions.setFlowOption(fileDescriptor, (SocketFlow)o);
            return;
        }
        else {
            if (Net.extendedOptions.isOptionSupported(socketOption)) {
                Net.extendedOptions.setOption(fileDescriptor, socketOption, o);
                return;
            }
            if (type != Integer.class && type != Boolean.class) {
                throw new AssertionError("Should not reach here");
            }
            if ((socketOption == StandardSocketOptions.SO_RCVBUF || socketOption == StandardSocketOptions.SO_SNDBUF) && (int)o < 0) {
                throw new IllegalArgumentException("Invalid send/receive buffer size");
            }
            if (socketOption == StandardSocketOptions.SO_LINGER) {
                int intValue = (int)o;
                if (intValue < 0) {
                    o = Integer.valueOf(-1);
                }
                if (intValue > 65535) {
                    o = Integer.valueOf(65535);
                }
            }
            if (socketOption == StandardSocketOptions.IP_TOS) {
                int intValue2 = (int)o;
                if (intValue2 < 0 || intValue2 > 255) {
                    throw new IllegalArgumentException("Invalid IP_TOS value");
                }
            }
            if (socketOption == StandardSocketOptions.IP_MULTICAST_TTL) {
                int intValue3 = (int)o;
                if (intValue3 < 0 || intValue3 > 255) {
                    throw new IllegalArgumentException("Invalid TTL/hop value");
                }
            }
            OptionKey option = SocketOptionRegistry.findOption(socketOption, protocolFamily);
            if (option == null) {
                throw new AssertionError("Option not found");
            }
            int intValue4;
            if (type == Integer.class) {
                intValue4 = (int)o;
            }
            else {
                intValue4 = (((boolean)o) ? 1 : 0);
            }
            setIntOption0(fileDescriptor, protocolFamily == Net.UNSPEC, option.level(), option.name(), intValue4, protocolFamily == StandardProtocolFamily.INET6);
            return;
        }
    }
    
    static Object getSocketOption(FileDescriptor fileDescriptor, ProtocolFamily protocolFamily, SocketOption<?> socketOption) throws IOException {
        Class<Object> type = socketOption.type();
        if (type == SocketFlow.class) {
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                securityManager.checkPermission(new NetworkPermission("getOption.SO_FLOW_SLA"));
            }
            SocketFlow create = SocketFlow.create();
            jdk.net.ExtendedSocketOptions.getFlowOption(fileDescriptor, create);
            return create;
        }
        else {
            if (Net.extendedOptions.isOptionSupported(socketOption)) {
                return Net.extendedOptions.getOption(fileDescriptor, socketOption);
            }
            if (type != Integer.class && type != Boolean.class) {
                throw new AssertionError("Should not reach here");
            }
            OptionKey option = SocketOptionRegistry.findOption(socketOption, protocolFamily);
            if (option == null) {
                throw new AssertionError("Option not found");
            }
            int intOption0 = getIntOption0(fileDescriptor, protocolFamily == Net.UNSPEC, option.level(), option.name());
            if (type == Integer.class) {
                return Integer.valueOf(intOption0);
            }
            return (intOption0 == 0) ? Boolean.FALSE : Boolean.TRUE;
        }
    }
    
    public static boolean isFastTcpLoopbackRequested() {
        String s = (String)AccessController.doPrivileged(new PrivilegedAction<String>() {
            @Override
            public String run() {
                return System.getProperty("jdk.net.useFastTcpLoopback");
            }
        });
        return "".equals(s) || Boolean.parseBoolean(s);
    }
    
    private static native boolean isIPv6Available0();
    
    private static native int isExclusiveBindAvailable();
    
    private static native boolean canIPv6SocketJoinIPv4Group0();
    
    private static native boolean canJoin6WithIPv4Group0();
    
    static FileDescriptor socket(boolean b) throws IOException {
        return socket(Net.UNSPEC, b);
    }
    
    static FileDescriptor socket(ProtocolFamily protocolFamily, boolean b) throws IOException {
        return IOUtil.newFD(socket0(isIPv6Available() && protocolFamily != StandardProtocolFamily.INET, b, false, Net.fastLoopback));
    }
    
    static FileDescriptor serverSocket(boolean b) {
        return IOUtil.newFD(socket0(isIPv6Available(), b, true, Net.fastLoopback));
    }
    
    private static native int socket0(boolean p0, boolean p1, boolean p2, boolean p3);
    
    public static void bind(FileDescriptor fileDescriptor, InetAddress inetAddress, int n) throws IOException {
        bind(Net.UNSPEC, fileDescriptor, inetAddress, n);
    }
    
    static void bind(ProtocolFamily protocolFamily, FileDescriptor fileDescriptor, InetAddress inetAddress, int n) throws IOException {
        bind0(fileDescriptor, isIPv6Available() && protocolFamily != StandardProtocolFamily.INET, Net.exclusiveBind, inetAddress, n);
    }
    
    private static native void bind0(FileDescriptor p0, boolean p1, boolean p2, InetAddress p3, int p4) throws IOException;
    
    static native void listen(FileDescriptor p0, int p1) throws IOException;
    
    static int connect(FileDescriptor fileDescriptor, InetAddress inetAddress, int n) throws IOException {
        return connect(Net.UNSPEC, fileDescriptor, inetAddress, n);
    }
    
    static int connect(ProtocolFamily protocolFamily, FileDescriptor fileDescriptor, InetAddress inetAddress, int n) throws IOException {
        return connect0(isIPv6Available() && protocolFamily != StandardProtocolFamily.INET, fileDescriptor, inetAddress, n);
    }
    
    private static native int connect0(boolean p0, FileDescriptor p1, InetAddress p2, int p3) throws IOException;
    
    static native void shutdown(FileDescriptor p0, int p1) throws IOException;
    
    private static native int localPort(FileDescriptor p0) throws IOException;
    
    private static native InetAddress localInetAddress(FileDescriptor p0) throws IOException;
    
    public static InetSocketAddress localAddress(FileDescriptor fileDescriptor) throws IOException {
        return new InetSocketAddress(localInetAddress(fileDescriptor), localPort(fileDescriptor));
    }
    
    private static native int remotePort(FileDescriptor p0) throws IOException;
    
    private static native InetAddress remoteInetAddress(FileDescriptor p0) throws IOException;
    
    static InetSocketAddress remoteAddress(FileDescriptor fileDescriptor) throws IOException {
        return new InetSocketAddress(remoteInetAddress(fileDescriptor), remotePort(fileDescriptor));
    }
    
    private static native int getIntOption0(FileDescriptor p0, boolean p1, int p2, int p3) throws IOException;
    
    private static native void setIntOption0(FileDescriptor p0, boolean p1, int p2, int p3, int p4, boolean p5) throws IOException;
    
    static native int poll(FileDescriptor p0, int p1, long p2) throws IOException;
    
    static int join4(FileDescriptor fileDescriptor, int n, int n2, int n3) throws IOException {
        return joinOrDrop4(true, fileDescriptor, n, n2, n3);
    }
    
    static void drop4(FileDescriptor fileDescriptor, int n, int n2, int n3) throws IOException {
        joinOrDrop4(false, fileDescriptor, n, n2, n3);
    }
    
    private static native int joinOrDrop4(boolean p0, FileDescriptor p1, int p2, int p3, int p4) throws IOException;
    
    static int block4(FileDescriptor fileDescriptor, int n, int n2, int n3) throws IOException {
        return blockOrUnblock4(true, fileDescriptor, n, n2, n3);
    }
    
    static void unblock4(FileDescriptor fileDescriptor, int n, int n2, int n3) throws IOException {
        blockOrUnblock4(false, fileDescriptor, n, n2, n3);
    }
    
    private static native int blockOrUnblock4(boolean p0, FileDescriptor p1, int p2, int p3, int p4) throws IOException;
    
    static int join6(FileDescriptor fileDescriptor, byte[] array, int n, byte[] array2) throws IOException {
        return joinOrDrop6(true, fileDescriptor, array, n, array2);
    }
    
    static void drop6(FileDescriptor fileDescriptor, byte[] array, int n, byte[] array2) throws IOException {
        joinOrDrop6(false, fileDescriptor, array, n, array2);
    }
    
    private static native int joinOrDrop6(boolean p0, FileDescriptor p1, byte[] p2, int p3, byte[] p4) throws IOException;
    
    static int block6(FileDescriptor fileDescriptor, byte[] array, int n, byte[] array2) throws IOException {
        return blockOrUnblock6(true, fileDescriptor, array, n, array2);
    }
    
    static void unblock6(FileDescriptor fileDescriptor, byte[] array, int n, byte[] array2) throws IOException {
        blockOrUnblock6(false, fileDescriptor, array, n, array2);
    }
    
    static native int blockOrUnblock6(boolean p0, FileDescriptor p1, byte[] p2, int p3, byte[] p4) throws IOException;
    
    static native void setInterface4(FileDescriptor p0, int p1) throws IOException;
    
    static native int getInterface4(FileDescriptor p0) throws IOException;
    
    static native void setInterface6(FileDescriptor p0, int p1) throws IOException;
    
    static native int getInterface6(FileDescriptor p0) throws IOException;
    
    private static native void initIDs();
    
    static native short pollinValue();
    
    static native short polloutValue();
    
    static native short pollerrValue();
    
    static native short pollhupValue();
    
    static native short pollnvalValue();
    
    static native short pollconnValue();
    
    static {
        UNSPEC = new ProtocolFamily() {
            @Override
            public String name() {
                return "UNSPEC";
            }
        };
        Net.checkedIPv6 = false;
        extendedOptions = ExtendedSocketOptions.getInstance();
        IOUtil.load();
        initIDs();
        POLLIN = pollinValue();
        POLLOUT = polloutValue();
        POLLERR = pollerrValue();
        POLLHUP = pollhupValue();
        POLLNVAL = pollnvalValue();
        POLLCONN = pollconnValue();
        int exclusiveBindAvailable = isExclusiveBindAvailable();
        if (exclusiveBindAvailable >= 0) {
            String s = (String)AccessController.doPrivileged(new PrivilegedAction<String>() {
                @Override
                public String run() {
                    return System.getProperty("sun.net.useExclusiveBind");
                }
            });
            if (s != null) {
                exclusiveBind = (s.length() == 0 || Boolean.parseBoolean(s));
            }
            else if (exclusiveBindAvailable == 1) {
                exclusiveBind = true;
            }
            else {
                exclusiveBind = false;
            }
        }
        else {
            exclusiveBind = false;
        }
        fastLoopback = isFastTcpLoopbackRequested();
    }
}
